"""
**ResponseGT**

A software package for computing electrical response properties of graph networks.

"""


# Project Details
__version__ = "1.1.4"
__install_version__ = "1.1.4"
__title__ = "ResponseGT"
__author__ = "William Stephenson and Dickson Owuor"
__credits__ = "The Regents of the University of Michigan"
